#include "Mesh.h"
#include <iostream.h>

/**
   parse input and create mesh.
   mesh encoded in a subset of the ray format.

   recognized structures:
     #vertex px py pz nx ny nz ts tt
       px py pz -- position of vertex
       nx ny nz -- ignored
       ts tt    -- texture coordinates

     #shape_triangle m i1 i2 i3
       m        -- ignored
       i1 i2 i3 -- indices of vertex

   all other structures are silently ignored.
 **/
bool
GetNextStructure (FILE *file, char *buffer)
{
  // skip to next string starting with '#', which is how we denote structures
  do
  {
    if(fscanf(file, "%s", buffer) != 1)
      return false;
  } while(*buffer != '#');

  return true;
}

Mesh::
Mesh (FILE *file)
{
  assert(file);
  int res;
  char buffer[1000];

  while(GetNextStructure(file, buffer))
  {
    if(!strcmp(buffer, "#vertex"))
    {
      // #vertex px py pz nx ny nz ts tt
      float px, py, pz, nx, ny, nz, ts, tt;

      res = fscanf(file, " %f %f %f %f %f %f %f %f",
        &px, &py, &pz, &nx, &ny, &nz, &ts, &tt);
      if(res != 8) {
        cerr << "parse error, vertex " << numV() << endl; exit(-1);
      }

      // add vertex and set position
      V.push_back(new Vertex());
      V.back()->getPos().set(px, py, pz);
    }

    else if(!strcmp(buffer, "#shape_triangle"))
    {
      // #shape_triangle m i1 i2 i3
      int m, i1, i2, i3;

      res = fscanf(file, " %d %d %d %d", &m, &i1, &i2, &i3);
      if(res != 4) {
        cerr << "parse error, face " << numF() << endl; exit(-1);
      }

      if(i1<0 || i1>=numV() || i2<0 || i2>=numV() || i3<0 || i3>=numV()) {
        cerr << "parse error, face " << numF() << " vertex indices out of bounds" << endl; exit(-1);
      }

      // add face and set connectivity
      F.push_back(new Face(getVertex(i1), getVertex(i2), getVertex(i3)));
    }
  }
}


/**
 * bounding box
 **/
void Mesh::
getBBox (Vector3f& min, Vector3f& max) const
{
#ifdef USE_UNIX
  std::vector< VertexPtr >::const_iterator v = V.begin();
#else
  const VertexPtr *v = V.begin();
#endif
  min = (*v)->getPos();
  max = (*v)->getPos();

  for(++v; v!=V.end(); ++v)
  {
    min = Min(min, (*v)->getPos());
    max = Max(max, (*v)->getPos());
  }
}


/**
 * write out mesh
 **/
bool Mesh::
write (FILE *file) const
{
  // write vertex_num
  if(fprintf(file, "#vertex_num %d\n", numV()) < 0) {
    cerr << "error writing to file" << endl; return false;
  }

  // write vertices
#ifdef USE_UNIX
  std::vector< VertexPtr >::const_iterator v;
#else
  const VertexPtr *v;
#endif
  for(v=V.begin(); v!=V.end(); ++v)
  {
    Vector3f p = (*v)->getPos();

    if(fprintf(file, "#vertex %f %f %f 0 0 0 0 0\n",  p.x,p.y,p.z) < 0) {
      cerr << "error writing to file" << endl; return false;
    }
  }

  // write triangles
#ifdef USE_UNIX
  std::vector< FacePtr >::const_iterator f;
#else
  const FacePtr *f;
#endif
  for(f=F.begin(); f!=F.end(); ++f)
  {
    int idx[3];

    // find vertex indices
    // if we assume that the vertices were allocated in one contiguous chunk
    // then this is a simple pointer subtraction, but to be safe we'll do a search
    for(int i=0; i<(*f)->val(); ++i)
    {
      const VertexPtr v = (*f)->getVertex(i);

      int j;

      for(j=0; j<numV(); ++j)
        if(getVertex(j) == v)
        { idx[i]=j; break; }

      assert(j < numV());
    }

    if(fprintf(file, "#shape_triangle -1 %d %d %d\n", idx[0], idx[1], idx[2]) < 0) {
      cerr << "error writing to file" << endl; return false;
    }
  }

  return true;
}


/**
 * openGL draw routines
 **/
void Mesh::
drawVertices () const
{
  // this draws all the vertices
  // optionally adds labels to them
#ifdef USE_UNIX
  std::vector< VertexPtr >::const_iterator v;
#else
  const VertexPtr *v;
#endif
  for(v=V.begin(); v!=V.end(); ++v)
  {
    glBegin(GL_POINTS);
    LoadVertex((*v)->getPos());
    glEnd();
  }
}

void Mesh::
drawEdges () const
{
  /**
   * NOTE: this draws all the edges twice,
   * can you think of a way to fix this?
   **/
  glBegin(GL_LINES);

#ifdef USE_UNIX
  std::vector< FacePtr >::const_iterator f;
#else
  const FacePtr *f;
#endif
  for(f=F.begin(); f!=F.end(); ++f)
  {
    for(int i=0; i<(*f)->val(); ++i)
    {
      LoadVertex((*f)->getVertex(i-1)->getPos());
      LoadVertex((*f)->getVertex(i+1)->getPos());
    }
  }

  glEnd();
}

void Mesh::
drawFaces () const
{
  glBegin(GL_TRIANGLES);

#ifdef USE_UNIX
  std::vector< FacePtr >::const_iterator f;
#else
  const FacePtr *f;
#endif
  for(f=F.begin(); f!=F.end(); ++f)
  {
    Vector3f
      p0 = (*f)->getVertex(0)->getPos(),
      p1 = (*f)->getVertex(1)->getPos(),
      p2 = (*f)->getVertex(2)->getPos();

    LoadNormal(Cross(p0-p1, p2-p1));
    LoadVertex(p0);
    LoadVertex(p1);
    LoadVertex(p2);
  }

  glEnd();
}

void Mesh::
drawFacesSmooth () const
{
  /** WORK HERE **/
  cerr << "draw smooth faces not implemented" << endl;
}
